﻿using System;
using System.IO.MemoryMappedFiles;
using System.Runtime.InteropServices;

namespace SeAH_Pipe_Tab_Cutter_Lib
{
    class EtherCAT
    {
        [StructLayout(LayoutKind.Explicit, Pack = 1)]
        public unsafe struct STRUCT_1
        {
            void Init()
            {
                this.Count = 0;
            }
            [FieldOffset(0)]
            public short Count;
            [FieldOffset(4)]
            public fixed byte value[1024];
        }

        /*************************
        ADS(EtherCAT Master) -> GUI(C#)
        1. Digital Input Data (1bit x 16ch x 2EA)
        2. Motor Feedback
          - actual position
          - actual velocity (maybe count per sec or ms)
          - actual cur (mA?)
          - driver state
          - driver error state
        3. SSI Encoder Interface (?bit x 5EA)
        4. Etc
        **************************/

        //1. Digital Input Data (1bit x 16ch x 2EA)
        [StructLayout(LayoutKind.Explicit, Pack = 1)]
        public unsafe struct ADS_TO_GUI_DIN
        {
            [FieldOffset(0)]
            public fixed sbyte data[16];
        }
        //2. analog input data (signed 16bit x 4ch x 2EA)
        [StructLayout(LayoutKind.Explicit, Pack = 1)]
        public unsafe struct ADS_TO_GUI_AIN
        {
            [FieldOffset(0)]
            public fixed ushort data[4];
        }
        //3. Motor Feedback
        //  - actual position		(4byte)
        //  - actual cur (0.1%/cnt)	(2byte)
        //  - driver state		(
        [StructLayout(LayoutKind.Explicit, Pack = 1)]
        public unsafe struct ADS_TO_GUI_MOTOR
        {
            [FieldOffset(0)]
            public int act_pos;
            [FieldOffset(4)]
            public short act_cur;
            [FieldOffset(6)]
            public ushort error_code;
            [FieldOffset(8)]
            public ushort driver_ready;
        }

        //4. SSI Encoder 32bit x 2ch x 3EA
        [StructLayout(LayoutKind.Explicit, Pack = 1)]
        public unsafe struct ADS_TO_GUI_SSI
        {
            [FieldOffset(0)]
            public fixed uint abs_encoder[2];
        };

        //4. etc (dummy) 
        [StructLayout(LayoutKind.Explicit, Pack = 1)]
        public unsafe struct ADS_TO_GUI_ETC
        {
            [FieldOffset(0)]
            public uint twincat_counter; // Twincat이 동작하는동안 무작정 증가시키는 카운터 (interval : 10ms)
            [FieldOffset(4)]
            public uint gui_counter;     // GUI가 구동되는동안 무조건적으로 증가시키는 카운터 (interval : 50ms)
            [FieldOffset(8)]
            public uint ads_counter;	 // ADS 통신을 한번 수행할때마다 증가시키는 카운터 (interval : 10ms)
        }

        [StructLayout(LayoutKind.Explicit, Pack = 1)]
        public unsafe struct ADS_TO_GUI
        {
            [FieldOffset(0)]
            public ADS_TO_GUI_DIN[] din;      // EL1809(16ch) x 2EA
            [FieldOffset(16 * 2)]
            public ADS_TO_GUI_AIN[] ain;      // EL3154(4ch) x 2EA
            [FieldOffset(16 * 2 + 4 * 2)]
            public ADS_TO_GUI_MOTOR[] motor;  // HIGEN Driver x 6EA
            [FieldOffset(16 * 2 + 4 * 2 + 10 * 6)]
            public ADS_TO_GUI_SSI[] ssi;      // EL5002(2ch) x 3EA
            [FieldOffset(16 * 2 + 4 * 2 + 10 * 6 +   4 * 3)]
            public ADS_TO_GUI_ETC etc;
            void Init()
            {
                din = new ADS_TO_GUI_DIN[2];
                ain = new ADS_TO_GUI_AIN[2];
                motor = new ADS_TO_GUI_MOTOR[6];
                ssi = new ADS_TO_GUI_SSI[3];
                etc = new ADS_TO_GUI_ETC();
            }
        }

        MemoryMappedFile mapFile = MemoryMappedFile.OpenExisting(
                        "SeAH_Pipe_Tab_SharedMemory",
                        MemoryMappedFileRights.ReadWrite);

        ADS_TO_GUI stdata = new ADS_TO_GUI();

        public int EtherCAT_MMFOpen()
        {
            //access memory mapped file (need persistence)
            int size = Marshal.SizeOf(typeof(ADS_TO_GUI));
            byte[] Buffer = new byte[size];

            var memMapFile = MemoryMappedFile.CreateOrOpen("SeAH_Pipe_Tab_SharedMemory", size, MemoryMappedFileAccess.Read);//, MemoryMappedFileOptions.None, CustomSecurity, System.IO.HandleInheritability.Inheritable))
            using (var accessor = memMapFile.CreateViewAccessor(0, size, MemoryMappedFileAccess.Read))
            {
                // 1. 에러 발생하여 2. 방법으로 시도
                //accessor.Read<ADS_TO_GUI>((long)0, out data);  // 참조가 사용이 되어 에러 발생.

                //accessor.Read(0, out data);

                for (long i = 0; i < size; i++)
                    Buffer[i] = accessor.ReadByte(i);
            }
            // 2. 버퍼를 사용을 하여 버퍼에서 구조체를 복사하는 방법 시도.
            ByteToStructure(Buffer, stdata.GetType());

            return 0;
        }

        //byte 배열을 구조체로
        public void ByteToStructure(byte[] data, Type type)
        {
            IntPtr buff = Marshal.AllocHGlobal(data.Length); // 배열의 크기만큼 비관리 메모리 영역에 메모리를 할당한다.
            Marshal.Copy(data, 0, buff, data.Length); // 배열에 저장된 데이터를 위에서 할당한 메모리 영역에 복사한다.
            //object obj = 
            stdata = (ADS_TO_GUI)Marshal.PtrToStructure(buff, typeof(ADS_TO_GUI)); // 복사된 데이터를 구조체 객체로 변환한다.--> 보호된 메모리를 읽거나 쓰려고 했습니다
            Marshal.FreeHGlobal(buff); // 비관리 메모리 영역에 할당했던 메모리를 해제함

            //if (Marshal.SizeOf(obj) != data.Length)// (((PACKET_DATA)obj).TotalBytes != data.Length) // 구조체와 원래의 데이터의 크기 비교
            //{
            //    return null; // 크기가 다르면 null 리턴
            //}
            //return obj; // 구조체 리턴
        }

    }
}
